@@ -26,8 +26,7 @@
  * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  * DEALINGS IN THE SOFTWARE.
  ****************************************************************************/
- 
- 
+
 #include "ogrgeopackageutility.h"
 #include "ogr_p.h"
 
@@ -38,16 +37,18 @@ OGRErr SQLCommand(sqlite3 * poDb, const char * pszSQL)
     CPLAssert( pszSQL != NULL );
 
     char *pszErrMsg = NULL;
+    //CPLDebug("GPKG", "exec(%s)", pszSQL);
     int rc = sqlite3_exec(poDb, pszSQL, NULL, NULL, &pszErrMsg);
-    
+
     if ( rc != SQLITE_OK )
     {
         CPLError( CE_Failure, CPLE_AppDefined,
                   "sqlite3_exec(%s) failed: %s",
-                  pszSQL, pszErrMsg );
+                  pszSQL, pszErrMsg ? pszErrMsg : "" );
+        sqlite3_free( pszErrMsg );
         return OGRERR_FAILURE;
     }
-    
+
     return OGRERR_NONE;
 }
 
@@ -73,14 +74,14 @@ OGRErr SQLQuery(sqlite3 * poDb, const char * pszSQL, SQLResult * poResult)
 
     poResult->rc = sqlite3_get_table(
         poDb, pszSQL,
-        &(poResult->papszResult), 
-        &(poResult->nRowCount), 
-        &(poResult->nColCount), 
+        &(poResult->papszResult),
+        &(poResult->nRowCount),
+        &(poResult->nColCount),
         &(poResult->pszErrMsg) );
-    
+
     if( poResult->rc != SQLITE_OK )
     {
-        CPLError( CE_Failure, CPLE_AppDefined, 
+        CPLError( CE_Failure, CPLE_AppDefined,
                   "sqlite3_get_table(%s) failed: %s", pszSQL, poResult->pszErrMsg );
         return OGRERR_FAILURE;
     }
@@ -96,52 +97,52 @@ OGRErr SQLResultFree(SQLResult * poResult)
 
     if ( poResult->pszErrMsg )
         sqlite3_free(poResult->pszErrMsg);
-        
+
     return OGRERR_NONE;
 }
 
 const char* SQLResultGetColumn(const SQLResult * poResult, int iColNum)
 {
-    if ( ! poResult ) 
+    if ( ! poResult )
         return NULL;
-        
+
     if ( iColNum < 0 || iColNum >= poResult->nColCount )
         return NULL;
-    
+
     return poResult->papszResult[iColNum];
 }
 
 const char* SQLResultGetValue(const SQLResult * poResult, int iColNum, int iRowNum)
 {
-    if ( ! poResult ) 
+    if ( ! poResult )
         return NULL;
 
     int nCols = poResult->nColCount;
-    int nRows = poResult->nRowCount;    
-        
+    int nRows = poResult->nRowCount;
+
     if ( iColNum < 0 || iColNum >= nCols )
         return NULL;
 
     if ( iRowNum < 0 || iRowNum >= nRows )
         return NULL;
-        
+
     return poResult->papszResult[ nCols + iRowNum * nCols + iColNum ];
 }
 
 int SQLResultGetValueAsInteger(const SQLResult * poResult, int iColNum, int iRowNum)
 {
-    if ( ! poResult ) 
+    if ( ! poResult )
         return 0;
-        
+
     int nCols = poResult->nColCount;
     int nRows = poResult->nRowCount;
-    
+
     if ( iColNum < 0 || iColNum >= nCols )
         return 0;
 
     if ( iRowNum < 0 || iRowNum >= nRows )
         return 0;
-    
+
     char *pszValue = poResult->papszResult[ nCols + iRowNum * nCols + iColNum ];
     if ( ! pszValue )
         return 0;
@@ -150,15 +151,16 @@ int SQLResultGetValueAsInteger(const SQLResult * poResult, int iColNum, int iRow
 }
 
 /* Returns the first row of first column of SQL as integer */
-int SQLGetInteger(sqlite3 * poDb, const char * pszSQL, OGRErr *err)
+GIntBig SQLGetInteger64(sqlite3 * poDb, const char * pszSQL, OGRErr *err)
 {
     CPLAssert( poDb != NULL );
-    
+
     sqlite3_stmt *poStmt;
-    int rc, i;
-    
+    int rc;
+    GIntBig i;
+
     /* Prepare the SQL */
-    rc = sqlite3_prepare_v2(poDb, pszSQL, strlen(pszSQL), &poStmt, NULL);
+    rc = sqlite3_prepare_v2(poDb, pszSQL, -1, &poStmt, NULL);
     if ( rc != SQLITE_OK )
     {
         CPLError( CE_Failure, CPLE_AppDefined, "sqlite3_prepare_v2(%s) failed: %s",
@@ -166,60 +168,27 @@ int SQLGetInteger(sqlite3 * poDb, const char * pszSQL, OGRErr *err)
         if ( err ) *err = OGRERR_FAILURE;
         return 0;
     }
-    
+
     /* Execute and fetch first row */
     rc = sqlite3_step(poStmt);
     if ( rc != SQLITE_ROW )
     {
         if ( err ) *err = OGRERR_FAILURE;
+        sqlite3_finalize(poStmt);
         return 0;
     }
-    
+
     /* Read the integer from the row */
-    i = sqlite3_column_int(poStmt, 0);
+    i = sqlite3_column_int64(poStmt, 0);
     sqlite3_finalize(poStmt);
-    
+
     if ( err ) *err = OGRERR_NONE;
     return i;
 }
 
-
-/* Requirement 20: A GeoPackage SHALL store feature table geometries */
-/* with the basic simple feature geometry types (Geometry, Point, */
-/* LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon, */
-/* GeomCollection) */
-/* http://opengis.github.io/geopackage/#geometry_types */
-OGRwkbGeometryType GPkgGeometryTypeToWKB(const char *pszGpkgType, int bHasZ)
+int SQLGetInteger(sqlite3 * poDb, const char * pszSQL, OGRErr *err)
 {
-    OGRwkbGeometryType oType;
-    
-    if ( EQUAL("Geometry", pszGpkgType) )
-        oType = wkbUnknown;
-    else if ( EQUAL("Point", pszGpkgType) )
-        oType =  wkbPoint;
-    else if ( EQUAL("LineString", pszGpkgType) )
-        oType =  wkbLineString;
-    else if ( EQUAL("Polygon", pszGpkgType) )
-        oType =  wkbPolygon;
-    else if ( EQUAL("MultiPoint", pszGpkgType) )
-        oType =  wkbMultiPoint;
-    else if ( EQUAL("MultiLineString", pszGpkgType) )
-        oType =  wkbMultiLineString;
-    else if ( EQUAL("MultiPolygon", pszGpkgType) )
-        oType =  wkbMultiPolygon;
-    else if ( EQUAL("GeometryCollection", pszGpkgType) )
-        oType =  wkbGeometryCollection;
-    else
-        oType =  wkbNone;
-
-    if ( (oType != wkbNone) && bHasZ )
-    {
-        unsigned int oi = oType;
-        oi &= wkb25DBit;
-        oType = (OGRwkbGeometryType)oi;
-    }
-
-    return oType;
+    return (int)SQLGetInteger64(poDb, pszSQL, err);
 }
 
 /* Requirement 20: A GeoPackage SHALL store feature table geometries */
@@ -227,73 +196,93 @@ OGRwkbGeometryType GPkgGeometryTypeToWKB(const char *pszGpkgType, int bHasZ)
 /* LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon, */
 /* GeomCollection) */
 /* http://opengis.github.io/geopackage/#geometry_types */
-const char* GPkgGeometryTypeFromWKB(OGRwkbGeometryType oType)
+OGRwkbGeometryType GPkgGeometryTypeToWKB(const char *pszGpkgType, bool bHasZ, bool bHasM)
 {
-    oType = wkbFlatten(oType);
-    
-    switch(oType)
+    OGRwkbGeometryType oType;
+
+    if ( EQUAL("Geometry", pszGpkgType) )
+        oType = wkbUnknown;
+    /* The 1.0 spec is not completely clear on what should be used... */
+    else if ( EQUAL("GeomCollection", pszGpkgType) ||
+              EQUAL("GeometryCollection", pszGpkgType) )
+        oType =  wkbGeometryCollection;
+    else
     {
-        case wkbPoint:
-            return "point";
-        case wkbLineString:
-            return "linestring";
-        case wkbPolygon:
-            return "polygon";
-        case wkbMultiPoint:
-            return "multipoint";
-        case wkbMultiLineString:
-            return "multilinestring";
-        case wkbMultiPolygon:
-            return "multipolygon";
-        case wkbGeometryCollection:
-            return "geometrycollection";
-        default:
-            return NULL;
+        oType = OGRFromOGCGeomType(pszGpkgType);
+        if( oType == wkbUnknown )
+            oType = wkbNone;
     }
+
+    if ( (oType != wkbNone) && bHasZ )
+    {
+        oType = wkbSetZ(oType);
+    }
+    if ( (oType != wkbNone) && bHasM )
+    {
+        oType = wkbSetM(oType);
+    }
+
+    return oType;
 }
 
 /* Requirement 5: The columns of tables in a GeoPackage SHALL only be */
 /* declared using one of the data types specified in table GeoPackage */
 /* Data Types. */
 /* http://opengis.github.io/geopackage/#table_column_data_types */
-OGRFieldType GPkgFieldToOGR(const char *pszGpkgType)
+OGRFieldType GPkgFieldToOGR(const char *pszGpkgType, OGRFieldSubType& eSubType,
+                            int& nMaxWidth)
 {
+    eSubType = OFSTNone;
+    nMaxWidth = 0;
+
     /* Integer types */
-    if ( STRNCASECMP("INTEGER", pszGpkgType, 7) == 0 )
+    if ( STRNCASECMP("INT", pszGpkgType, 3) == 0 )
+        return OFTInteger64;
+    else if ( EQUAL("MEDIUMINT", pszGpkgType) )
         return OFTInteger;
-    else if ( STRNCASECMP("INT", pszGpkgType, 3) == 0 )
+    else if ( EQUAL("SMALLINT", pszGpkgType) )
+    {
+        eSubType = OFSTInt16;
         return OFTInteger;
-    else if ( STRNCASECMP("MEDIUMINT", pszGpkgType, 9) == 0 )
+    }
+    else if ( EQUAL("TINYINT", pszGpkgType) )
         return OFTInteger;
-    else if ( STRNCASECMP("SMALLINT", pszGpkgType, 8) == 0 )
-        return OFTInteger;
-    else if ( STRNCASECMP("TINYINT", pszGpkgType, 7) == 0 )
-        return OFTInteger;
-    else if ( STRNCASECMP("BOOLEAN", pszGpkgType, 7) == 0 )
+    else if ( EQUAL("BOOLEAN", pszGpkgType) )
+    {
+        eSubType = OFSTBoolean;
         return OFTInteger;
+    }
 
     /* Real types */
-    else if ( STRNCASECMP("FLOAT", pszGpkgType, 5) == 0 )
+    else if ( EQUAL("FLOAT", pszGpkgType) )
+    {
+        eSubType = OFSTFloat32;
         return OFTReal;
-    else if ( STRNCASECMP("DOUBLE", pszGpkgType, 6) == 0 )
+    }
+    else if ( EQUAL("DOUBLE", pszGpkgType) )
         return OFTReal;
-    else if ( STRNCASECMP("REAL", pszGpkgType, 4) == 0 )
+    else if ( EQUAL("REAL", pszGpkgType) )
         return OFTReal;
-        
+
     /* String/binary types */
     else if ( STRNCASECMP("TEXT", pszGpkgType, 4) == 0 )
+    {
+        if( pszGpkgType[4] == '(' )
+            nMaxWidth = atoi(pszGpkgType+5);
         return OFTString;
+    }
+
     else if ( STRNCASECMP("BLOB", pszGpkgType, 4) == 0 )
         return OFTBinary;
-        
+
     /* Date types */
-    else if ( STRNCASECMP("DATE", pszGpkgType, 4) == 0 )
+    else if ( EQUAL("DATE", pszGpkgType) )
         return OFTDate;
-    else if ( STRNCASECMP("DATETIME", pszGpkgType, 8) == 0 )
+    else if ( EQUAL("DATETIME", pszGpkgType) )
         return OFTDateTime;
 
     /* Illegal! */
-    else 
+    else
         return (OGRFieldType)(OFTMaxType + 1);
 }
 
@@ -301,16 +290,36 @@ OGRFieldType GPkgFieldToOGR(const char *pszGpkgType)
 /* declared using one of the data types specified in table GeoPackage */
 /* Data Types. */
 /* http://opengis.github.io/geopackage/#table_column_data_types */
-const char* GPkgFieldFromOGR(OGRFieldType nType)
+const char* GPkgFieldFromOGR(OGRFieldType nType, OGRFieldSubType eSubType,
+                             int nMaxWidth)
 {
     switch(nType)
     {
         case OFTInteger:
+        {
+            if( eSubType == OFSTBoolean )
+                return "BOOLEAN";
+            else if( eSubType == OFSTInt16 )
+                return "SMALLINT";
+            else
+                return "MEDIUMINT";
+        }
+        case OFTInteger64:
             return "INTEGER";
         case OFTReal:
-            return "REAL";
+        {
+            if( eSubType == OFSTFloat32 )
+                return "FLOAT";
+            else
+                return "REAL";
+        }
         case OFTString:
-            return "TEXT";
+        {
+            if( nMaxWidth > 0 )
+                return CPLSPrintf("TEXT(%d)", nMaxWidth);
+            else
+                return "TEXT";
+        }
         case OFTBinary:
             return "BLOB";
         case OFTDate:
@@ -318,7 +327,7 @@ const char* GPkgFieldFromOGR(OGRFieldType nType)
         case OFTDateTime:
             return "DATETIME";
         default:
-            return NULL;
+            return "TEXT";
     }
 }
 
@@ -347,24 +356,24 @@ int SQLiteFieldFromOGR(OGRFieldType nType)
 
 
 
-/* Requirement 19: A GeoPackage SHALL store feature table geometries 
-*  with or without optional elevation (Z) and/or measure (M) values in SQL 
-*  BLOBs using the Standard GeoPackageBinary format specified in table GeoPackage 
+/* Requirement 19: A GeoPackage SHALL store feature table geometries
+*  with or without optional elevation (Z) and/or measure (M) values in SQL
+*  BLOBs using the Standard GeoPackageBinary format specified in table GeoPackage
 *  SQL Geometry Binary Format and clause Geometry Encoding.
 *
 *  http://opengis.github.io/geopackage/#gpb_format
 *
 *   GeoPackageBinaryHeader {
-*     byte[2] magic = 0x4750; 
-*     byte version;           
-*     byte flags;             
+*     byte[2] magic = 0x4750;
+*     byte version;
+*     byte flags;
 *     int32 srs_id;
-*     double[] envelope;      
+*     double[] envelope;
 *    }
-* 
+*
 *   StandardGeoPackageBinary {
-*     GeoPackageBinaryHeader header; 
-*     WKBGeometry geometry;          
+*     GeoPackageBinaryHeader header;
+*     WKBGeometry geometry;
 *   }
 *
 *  Flags byte contents:
@@ -380,7 +389,7 @@ int SQLiteFieldFromOGR(OGRFieldType nType)
 GByte* GPkgGeometryFromOGR(const OGRGeometry *poGeometry, int iSrsId, size_t *pszWkb)
 {
     CPLAssert( poGeometry != NULL );
-    
+
     GByte *pabyPtr;
     GByte byFlags = 0;
     GByte byEnv = 1;
@@ -388,43 +397,45 @@ GByte* GPkgGeometryFromOGR(const OGRGeometry *poGeometry, int iSrsId, size_t *ps
     OGRErr err;
     OGRBoolean bPoint = (wkbFlatten(poGeometry->getGeometryType()) == wkbPoint);
     OGRBoolean bEmpty = poGeometry->IsEmpty();
+    /* We voluntarily use getCoordinateDimension() so as to get only 2 for XY/XYM */
+    /* and 3 for XYZ/XYZM as we currently don't write envelopes with M extent. */
     int iDims = poGeometry->getCoordinateDimension();
 
     /* Header has 8 bytes for sure, and optional extra space for bounds */
-    size_t szHeader = 2+1+1+4;    
+    size_t szHeader = 2+1+1+4;
     if ( ! bPoint && ! bEmpty )
     {
         szHeader += 8*2*iDims;
     }
-    
+
     /* Total BLOB size is header + WKB size */
     size_t szWkb = szHeader + poGeometry->WkbSize();
     GByte *pabyWkb = (GByte *)CPLMalloc(szWkb);
     if (pszWkb)
         *pszWkb = szWkb;
-    
+
     /* Header Magic */
     pabyWkb[0] = 0x47;
     pabyWkb[1] = 0x50;
 
     /* GPKG BLOB Version */
     pabyWkb[2] = 0;
-    
+
     /* Extended? No. */
 
     /* Envelope dimensionality? */
 
     /* Don't write envelope for point type */
     if ( bPoint )
-        byEnv = 0; 
+        byEnv = 0;
     else
-        /* 3D envelope for 3D data */      
+        /* 3D envelope for 3D data */
         if ( iDims == 3 )
             byEnv = 2;
-        /* 2D envelope otherwise */      
+        /* 2D envelope otherwise */
         else
             byEnv = 1;
-    
+
     /* Empty? No envelope then. */
     if ( bEmpty )
     {
@@ -432,7 +443,7 @@ GByte* GPkgGeometryFromOGR(const OGRGeometry *poGeometry, int iSrsId, size_t *ps
         /* Set empty flag */
         byFlags |= (1 << 4);
     }
-    
+
     /* Set envelope flags */
     byFlags |= (byEnv << 1);
 
@@ -442,10 +453,10 @@ GByte* GPkgGeometryFromOGR(const OGRGeometry *poGeometry, int iSrsId, size_t *ps
 
     /* Write flags byte */
     pabyWkb[3] = byFlags;
-    
+
     /* Write srs_id */
     memcpy(pabyWkb+4, &iSrsId, 4);
-    
+
     /* Write envelope */
     if ( ! bEmpty && ! bPoint )
     {
@@ -471,9 +482,9 @@ GByte* GPkgGeometryFromOGR(const OGRGeometry *poGeometry, int iSrsId, size_t *ps
             padPtr[3] = oEnv.MaxY;
         }
     }
-    
+
     pabyPtr = pabyWkb + szHeader;
-    
+
     /* Use the wkbVariantIso for ISO SQL/MM output (differs for 3d geometry) */
     err = poGeometry->exportToWkb(eByteOrder, pabyPtr, wkbVariantIso);
     if ( err != OGRERR_NONE )
@@ -481,18 +492,19 @@ GByte* GPkgGeometryFromOGR(const OGRGeometry *poGeometry, int iSrsId, size_t *ps
         CPLFree(pabyWkb);
         return NULL;
     }
-    
-    return pabyWkb; 
+
+    return pabyWkb;
 }
 
 
-OGRErr GPkgHeaderFromWKB(const GByte *pabyGpkg, GPkgHeader *poHeader)
+OGRErr GPkgHeaderFromWKB(const GByte *pabyGpkg, size_t szGpkg, GPkgHeader *poHeader)
 {
     CPLAssert( pabyGpkg != NULL );
     CPLAssert( poHeader != NULL );
 
     /* Magic (match required) */
-    if ( pabyGpkg[0] != 0x47 || 
+    if ( szGpkg < 8 ||
+         pabyGpkg[0] != 0x47 ||
          pabyGpkg[1] != 0x50 ||
          pabyGpkg[2] != 0 )  /* Version (only 0 supported at this time)*/
     {
@@ -504,29 +516,63 @@ OGRErr GPkgHeaderFromWKB(const GByte *pabyGpkg, GPkgHeader *poHeader)
     poHeader->bEmpty = (byFlags & (0x01 << 4)) >> 4;
     poHeader->bExtended = (byFlags & (0x01 << 5)) >> 5;
     poHeader->eByteOrder = (OGRwkbByteOrder)(byFlags & 0x01);
+    poHeader->bExtentHasXY = false;
+    poHeader->bExtentHasZ = false;
+#ifdef notdef
+    poHeader->bExtentHasM = false;
+#endif
     OGRBoolean bSwap = OGR_SWAP(poHeader->eByteOrder);
 
     /* Envelope */
     int iEnvelope = (byFlags & (0x07 << 1)) >> 1;
-    if ( iEnvelope == 1 )
-        poHeader->iDims = 2; /* 2D envelope */
-    else if ( iEnvelope == 2 )
-        poHeader->iDims = 3; /* 3D envelope */
-    else 
-        poHeader->iDims = 0; /* No envelope */
+    int nEnvelopeDim = 0;
+    if( iEnvelope )
+    {
+        poHeader->bExtentHasXY = true;
+        if( iEnvelope == 1 )
+        {
+            nEnvelopeDim = 2; /* 2D envelope */
+        }
+        else if ( iEnvelope == 2 )
+        {
+            poHeader->bExtentHasZ = true;
+            nEnvelopeDim = 3; /* 2D+Z envelope */
+        }
+        else if ( iEnvelope == 3 )
+        {
+#ifdef notdef
+            poHeader->bExtentHasM = true;
+#endif
+            nEnvelopeDim = 3; /* 2D+M envelope */
+        }
+        else if ( iEnvelope == 4 )
+        {
+            poHeader->bExtentHasZ = true;
+#ifdef notdef
+            poHeader->bExtentHasM = true;
+#endif
+            nEnvelopeDim = 4; /* 2D+ZM envelope */
+        }
+    }
 
     /* SrsId */
     int iSrsId;
-    memcpy(&iSrsId, pabyGpkg+4, 4);    
+    memcpy(&iSrsId, pabyGpkg+4, 4);
     if ( bSwap )
     {
         iSrsId = CPL_SWAP32(iSrsId);
     }
     poHeader->iSrsId = iSrsId;
-    
+
+    if( szGpkg < static_cast<size_t>(8 + 8*2*nEnvelopeDim) )
+    {
+        // Not enough bytes
+        return OGRERR_FAILURE;
+    }
+
     /* Envelope */
     double *padPtr = (double*)(pabyGpkg+8);
-    if ( poHeader->iDims == 2 )
+    if ( poHeader->bExtentHasXY )
     {
         poHeader->MinX = padPtr[0];
         poHeader->MaxX = padPtr[1];
@@ -540,7 +586,7 @@ OGRErr GPkgHeaderFromWKB(const GByte *pabyGpkg, GPkgHeader *poHeader)
             CPL_SWAPDOUBLE(&(poHeader->MaxY));
         }
     }
-    if ( poHeader->iDims == 3 )
+    if ( poHeader->bExtentHasZ )
     {
         poHeader->MinZ = padPtr[4];
         poHeader->MaxZ = padPtr[5];
@@ -550,31 +596,44 @@ OGRErr GPkgHeaderFromWKB(const GByte *pabyGpkg, GPkgHeader *poHeader)
             CPL_SWAPDOUBLE(&(poHeader->MaxZ));
         }
     }
-    
+#ifdef notdef
+    if ( poHeader->bExtentHasM )
+    {
+        poHeader->MinM = padPtr[ ( poHeader->bExtentHasZ ) ? 6 : 4 ];
+        poHeader->MaxM = padPtr[ ( poHeader->bExtentHasZ ) ? 7 : 5 ];
+        if ( bSwap )
+        {
+            CPL_SWAPDOUBLE(&(poHeader->MinM));
+            CPL_SWAPDOUBLE(&(poHeader->MaxM));
+        }
+    }
+#endif
+
     /* Header size in byte stream */
-    poHeader->szHeader = 8 + 8*2*(poHeader->iDims);
-    
+    poHeader->szHeader = 8 + 8*2*nEnvelopeDim;
+
     return OGRERR_NONE;
 }
 
-OGRGeometry* GPkgGeometryToOGR(GByte *pabyGpkg, size_t szGpkg, OGRSpatialReference *poSrs)
+OGRGeometry* GPkgGeometryToOGR(const GByte *pabyGpkg, size_t szGpkg, OGRSpatialReference *poSrs)
 {
     CPLAssert( pabyGpkg != NULL );
-    
+
     GPkgHeader oHeader;
     OGRGeometry *poGeom;
-    
+
     /* Read header */
-    OGRErr err = GPkgHeaderFromWKB(pabyGpkg, &oHeader);
+    OGRErr err = GPkgHeaderFromWKB(pabyGpkg, szGpkg, &oHeader);
     if ( err != OGRERR_NONE )
         return NULL;
 
     /* WKB pointer */
-    GByte *pabyWkb = pabyGpkg + oHeader.szHeader;
+    const GByte *pabyWkb = pabyGpkg + oHeader.szHeader;
     size_t szWkb = szGpkg - oHeader.szHeader;
 
     /* Parse WKB */
-    err = OGRGeometryFactory::createFromWkb(pabyWkb, poSrs, &poGeom, szWkb);
+    err = OGRGeometryFactory::createFromWkb((GByte*)pabyWkb, poSrs, &poGeom,
+                                            static_cast<int>(szWkb));
     if ( err != OGRERR_NONE )
         return NULL;
 
@@ -582,33 +641,29 @@ OGRGeometry* GPkgGeometryToOGR(GByte *pabyGpkg, size_t szGpkg, OGRSpatialReferen
 }
 
 
-OGRErr GPkgEnvelopeToOGR(GByte *pabyGpkg, CPL_UNUSED size_t szGpkg, OGREnvelope *poEnv)
+OGRErr GPkgEnvelopeToOGR(GByte *pabyGpkg,
+                         size_t szGpkg,
+                         OGREnvelope *poEnv)
 {
     CPLAssert( poEnv != NULL );
     CPLAssert( pabyGpkg != NULL );
-    
+
     GPkgHeader oHeader;
 
     /* Read header */
-    OGRErr err = GPkgHeaderFromWKB(pabyGpkg, &oHeader);
+    OGRErr err = GPkgHeaderFromWKB(pabyGpkg, szGpkg, &oHeader);
     if ( err != OGRERR_NONE )
         return err;
 
-    if ( oHeader.bEmpty || oHeader.iDims == 0 )
+    if ( oHeader.bEmpty || !oHeader.bExtentHasXY )
     {
         return OGRERR_FAILURE;
     }
-    
+
     poEnv->MinX = oHeader.MinX;
     poEnv->MaxX = oHeader.MaxX;
     poEnv->MinY = oHeader.MinY;
     poEnv->MaxY = oHeader.MaxY;
-    
-    // if ( oHeader.iDims == 3 )
-    // {
-    //     poEnv->MinZ = oHeader.MinZ;
-    //     poEnv->MaxZ = oHeader.MaxZ;
-    // }
-    
+
     return OGRERR_NONE;
 }
